/**
 * \brief Collection of utility functions
 * \author Edward Lu
 * \date March 20 2013
 */

#ifndef __UTILITY_H
#define __UTILITY_H

#include <SFML/System.hpp>

#include <random>

//The constant PI, in radians
#define PI 3.14159f
//http://www.tauday.com/tau-manifesto
#define TAU 6.283185f
//Conversion factor from radians to degrees
#define TO_DEG 57.2958279f
//Conversion factor from degrees to radians
#define TO_RAD 0.0174532778f
//Tolerance used in some calculations
#define EPSILON 0.01f

/** Shifts the angle to between 0 and 2*PI */
float shiftAngle(float angle);
/** Initializes the random seed engine that getRandom() uses. */
void initRandom();
/** Gets a random number in the range [0, 1]. */
float getRandom();
/** Gets a random number in the range [begin, end]. */
float getRandom(float begin, float end);
/** Checks whether the string is a valid property from GetLiteralProperty() */
bool IsValidProperty(const std::string& prop);
/** Treating p1 as the origin, returns the quadrant p2 is in. */
int getQuadrant(const sf::Vector2f &p1, const sf::Vector2f &p2);
/** Get "position" of the quadrant */
sf::Vector2f getQuadrantVec(int quadrant, float magnitude);
/** Returns closest distance from point to line */
float lineDistance(sf::Vector2f point, sf::Vector2f p1, sf::Vector2f p2);
/** Returns closest distance between non-intersecting lines */
float lineDistance(sf::Vector2f p1, sf::Vector2f p2, sf::Vector2f p3, sf::Vector2f p4);
/** Returns true if the point is (close to) on the line. */
bool onLine(sf::Vector2f p1, sf::Vector2f p2, sf::Vector2f point, float epsilon = EPSILON);
/** Returns true if the point is contained in the polygon. */
bool contains(const std::vector<sf::Vector2f> &poly, sf::Vector2f point);
/** Returns angle from base to vec, taking mid as origin */
float angleBetween(sf::Vector2f base, sf::Vector2f vec, sf::Vector2f mid);
/** Returns angle from base to vec */
float angleBetween(sf::Vector2f base, sf::Vector2f vec);

/** Returns the intersection point of two lines defined by two points each */
template<typename T>
bool lineIntersect(sf::Vector2<T> p1, sf::Vector2<T> p2, sf::Vector2<T> p3, sf::Vector2<T> p4, sf::Vector2<T> &intersection) {
    //Ran into some precision issues a while back
	double a1 = p2.y - p1.y;
	double b1 = p1.x - p2.x;
	double c1 = a1*p1.x + b1*p1.y;
	double a2 = p4.y - p3.y;
	double b2 = p3.x - p4.x;
	double c2 = a2*p3.x + b2*p3.y;
	double det = a1 * b2 - b1 * a2;

	if(det == 0) {
        return false;
	} else {
		intersection = sf::Vector2<T>((T)(b2*c1 - b1*c2)/det, (T)(a1*c2 - a2*c1)/det);
        return true;
	}
}

/** Returns dot product */
template <typename T>
T dot(const sf::Vector2<T> &v1, const sf::Vector2<T> &v2) {
	return v1.x * v2.x + v1.y * v2.y;
}

/** Gets the magnitude of vec */
template <typename T>
float magnitude(const sf::Vector2<T> vec) {
	return sqrt(vec.x*vec.x+vec.y*vec.y);
}

/** Returns normalized vector */
template <typename T>
sf::Vector2<T> normalize(const sf::Vector2<T> &vec) {
	if(vec.x == 0 && vec.y == 0) {
		return vec;
	} else {
		return vec / magnitude(vec);
	}
}

/** Returns the sign (-1, 0, 1) of a number. */
template <typename T>
int sign(T val) {
	return (T(0) < val) - (val < T(0));
}

template <typename T, typename R>
bool close(T a, R b) {
	return std::abs(a - b) <= EPSILON;
}

template <typename T, typename R>
bool close(sf::Vector2<T> v1, sf::Vector2<R> v2) {
	return close(v1.x, v2.x) && close(v1.y, v2.y);
}

template <typename T>
sf::Vector2i getNormal(sf::Vector2<T> p1, sf::Vector2<T> p2) {
	return sf::Vector2i(p1.y, p2.x) - sf::Vector2i(p2.y, p1.x);
}

#endif
