
#ifdef LINUX
#include <float.h>
#endif

#include <unordered_map>

#include "Level.h"

#include "Utility/Utility.h"
#include "Entity.h"
#include "Activatable.h"
#include "Movable.h"
#include "Player.h"
#include "Demon.h"

Level::Level(Player& player)
	: player(player)
{
	mapSize = sf::Vector2u(0, 0);
	tileSize = sf::Vector2u(0, 0);
	isActive = false;
	HUDText = "";
}

Level::Level(Player& player, const std::string& mapName, const std::string& spawnName)
	: player(player)
{
	mapSize = sf::Vector2u(0, 0);
	tileSize = sf::Vector2u(0, 0);
	isActive = false;

	LoadMap(mapName, spawnName);
}

bool Level::LoadMap(const std::string& mapName, const std::string& spawnName)
{
	if(!Parse(mapName)) {
		return false;
	}

	//Spawn the player
	SpawnPlayer(spawnName);

	//The level is now active and can be used.
	isActive = true;
	
	return true;
}

void Level::SpawnPlayer(const std::string& spawnName)
{
	std::map<std::string, sf::Vector2f>::iterator spawnIter;
	spawnIter = spawns.find(spawnName);
	if(spawnIter != spawns.end()) {
		player.SetPos(spawnIter->second);
	} else {
		printf("WARNING: Spawnpoint %s not found! Defaulting to (0, 0)...\n", spawnName.c_str());
		player.SetPos(sf::Vector2f(0, 0));
	}
}

void Level::SetActive(bool active)
{
	isActive = active;
	//If the level was reactivated, we want to display a message to the player
	if(active) {
		SetHUDText("Now in " + description + ".");
	}
}

void Level::UnloadMap()
{
	//Destroy all pointer vectors
	for(unsigned int i = 0; i < activatables.size(); i++)
		delete(activatables[i]);
	activatables.clear();
	for(unsigned int i = 0; i < enemies.size(); i++)
		delete(enemies[i]);
	enemies.clear();
	//Remove all attack cones
	attackCones.clear();
	bullets.clear();
}

Level::~Level()
{
	//Destroy all pointer vectors
	for(unsigned int i = 0; i < activatables.size(); i++)
		delete(activatables[i]);
	activatables.clear();
	for(unsigned int i = 0; i < enemies.size(); i++)
		delete(enemies[i]);
}

bool Level::Parse(const std::string& mapName)
{
	Tmx::Map map;

	printf("Parsing map...\n");
	//Attempt to parse the map, and report any errors
	map.ParseFile(BASEMAPDIR + mapName);

	//Initialize some properties
	mapSize = sf::Vector2u(map.GetWidth(), map.GetHeight());
	tileSize = sf::Vector2u(map.GetTileWidth(), map.GetTileHeight());
	
	if(map.HasError()) {
		fprintf(stderr, "WARNING: Could not parse level %s: %s\n",
			mapName.c_str(), map.GetErrorText().c_str());
		return false;
	}
	
	//Load the map into the display object
	tilemap.LoadMap(map);
    
    //Load the map into the collision object
    mapCollider.LoadMap(map);

	//Parse the map objects
	for (int i = 0; i < map.GetNumObjectGroups(); ++i) {
		//Get an object group.
		const Tmx::ObjectGroup *objectGroup = map.GetObjectGroup(i);
        bool collision = objectGroup->GetProperties().GetIntProperty("Collision");

        //If this is a collision group, just skip it
        if(collision) {
            continue;
        }

		//Iterate through all objects in the collision object group
		for (int j = 0; j < objectGroup->GetNumObjects(); ++j) {
			//Grab objects from the group
			const Tmx::Object *object = objectGroup->GetObject(j);

			if(!LoadObject(object)) {
				fprintf(stderr, "WARNING: Found unknown object type %s\n", object->GetType().c_str());
			}
		}
	}

	//Get a map description to display to the player.
	description = map.GetProperties().GetStringProperty("Description");
	printf("Map parsing complete\n");
	return true;
}

bool Level::LoadObject(const Tmx::Object *object) {
	const std::string &type = object->GetType();
	sf::Vector2f objectPos((float)object->GetX() + object->GetWidth()/2.0f,
				(float)object->GetY() - object->GetHeight()/2.0f);

	//Check the objects against our known object types...
	//Is the object a (the) spawn?
	if(type.compare("PlayerSpawn") == 0) {
		//If the spawn has no name, we can't store it
		std::string spawnName = object->GetName();
		if(spawnName.empty()) {
			fprintf(stderr, "WARNING: Spawn at %d, %d has no name! Ignoring...\n",
				object->GetX(), object->GetY());
		} else {
			spawns.insert(std::pair<std::string, sf::Vector2f>(spawnName, objectPos));
		}
	} else if(type.compare("Demon") == 0) {
		enemies.push_back(new Demon(objectPos));
	}
	//We don't know what type of object this is, issue an error
	else {
		return false;
	}
	return true;
}

sf::Vector2i Level::GetGlobalTile(const sf::Vector2f& pos) const
{
	return sf::Vector2i(std::lround(pos.x) / tileSize.x, std::lround(pos.y) / tileSize.y);
}

sf::Vector2i Level::GetPixel(const sf::Vector2f& pos) const
{
	return sf::Vector2i(std::lround(pos.x) % tileSize.x, std::lround(pos.y) % tileSize.y);
}

void Level::DoActivate()
{
	for(unsigned int i = 0; i < activatables.size(); i++) {
		/*if(activatables[i]->IsColliding(player)) {
			activatables[i]->Activate();
			didActivate = true;
		}*/
	}
}

void Level::AddAttack(const AttackCone& attack)
{
	attackCones.push_back(attack);
}

void Level::AddBullet(Bullet &&bullet) {
	bullets.push_back(bullet);
}

sf::Vector2i Level::GetSize() const
{
	return sf::Vector2i(mapSize.x*tileSize.x, mapSize.y*tileSize.y);
}

Player& Level::GetPlayer()
{
	return player;
}

const Player& Level::GetPlayer() const
{
	return player;
}

bool Level::GetCollideAngle(sf::Vector2f point, float angle, sf::Vector2f &closestIntersect, int mask) const {
	//Make the algorithm a bit more stable by setting the addition to a fixed length
	sf::Vector2f dir = sf::Vector2f(50.0f * cos(angle), 50.0f * sin(angle));
	return GetCollide(point, point + dir, closestIntersect, mask);
}

bool Level::GetCollideDir(sf::Vector2f point, sf::Vector2f dir, sf::Vector2f &closestIntersect, int mask) const {
	//Make the algorithm a bit more stable by setting the addition to a fixed length
	float dirMag = magnitude(dir);
	dir = dir/dirMag * 50.0f;
	return GetCollide(point, point + dir, closestIntersect, mask);
}

bool Level::GetCollide(sf::Vector2f p1, sf::Vector2f p2, sf::Vector2f& closestIntersect, int mask) const {
	float closestMag = FLT_MAX;
	sf::Vector2f dir = p2 - p1;

    if(mask & COLLIDE_WALLS) {
        bool found = mapCollider.GetCollide(p1, p2, closestIntersect);
        if(found) {
            closestMag = magnitude(p1 - closestIntersect);
        }
    }

	for(unsigned j = 0; j < activatables.size() + enemies.size(); j++) {
		Entity *entity;
		if(j < activatables.size()) {
			if(!activatables[j]->IsCollidable() || !(mask & COLLIDE_ACTIVATABLES)) { continue; }
			entity = activatables[j];
		} else {
            int i = j - activatables.size();
			if(!enemies[i]->IsCollidable() || !enemies[i]->IsAlive() || !(mask & COLLIDE_ENEMIES)) { continue; }
			entity = enemies[i];
		}

		sf::Vector2f intersect = entity->lineIntersect(p1, p2, EPSILON);
		float intersectMag = magnitude(p1 - intersect);

		if(intersectMag < closestMag &&
				sign(intersect.x - p1.x) == sign(dir.x) &&
				sign(intersect.y - p1.y) == sign(dir.y)) {
			closestIntersect = intersect;
			closestMag = intersectMag;
		}
	}

	return closestMag != FLT_MAX;
}

void Level::Update(const sf::Time& timePassed)
{
	for(unsigned int i = 0; i < activatables.size(); i++) {
		activatables[i]->Update(*this, timePassed);
		//Check for stairs, the only activatable that can be both active and finished
		/*if(activatables[i]->IsActive() && activatables[i]->IsFinished()) {
			//We've got to notify the Game that this level is done and we'd like to
			// transition to another.
			isActive = false;
		}*/
	}

	for(unsigned int i = 0; i < enemies.size(); i++) {
		enemies[i]->Update(*this, timePassed);
		/*if(player.IsColliding(*enemies[i])) {
			enemies[i]->Attack(player);
			//printf("Player colliding with enemy %d\n", i);
		}*/
	}

	auto bulletIter = bullets.begin();
	while(bulletIter != bullets.end()) {
		bulletIter->Update(*this, timePassed);
		if(!bulletIter->IsActive()) {
			auto temp = bulletIter;
			++bulletIter;
			bullets.erase(temp);
		} else {
			//Check for enemy collsions
			for(int i = 0; i < enemies.size(); i++) {
				Unit *enemy = enemies[i];
				if(bulletIter->IsColliding(*enemy)) {
					bulletIter->Damage(*enemy);
				}
			}
			++bulletIter;
		}
	}

	//Iterate through all the player's attacks
	std::list<AttackCone>::iterator p = attackCones.begin();
	while(p != attackCones.end()) {
		//Grab p's current value into a temporary value and then increment p
		//We do this so that we delete the temp value, not p itself
		std::list<AttackCone>::iterator temp = p++;
		//Update the attack box
		temp->Update(*this, timePassed);
		if(temp->IsExpired()) {
			//If it's now expired, destroy it
			attackCones.erase(temp);
		} else if(&temp->GetOwner() == &player) {
			//Otherwise, check for collisions with enemies
			for(unsigned int i = 0; i < enemies.size(); i++) {
				if(!temp->IsUnitHit(*enemies[i]) && enemies[i]->IsColliding(*temp)) {
					temp->GetOwner().Attack(*enemies[i]);
					temp->UnitHit(*enemies[i]);
				}
			}
		} else {
			if(!temp->IsUnitHit(player) && player.IsColliding(*temp)) {
				temp->GetOwner().Attack(player);
				temp->UnitHit(player);
			}
		}
	}

	player.Update(*this, timePassed);
}

void Level::SetHUDText(const std::string text)
{
	HUDText = text;
}

void Level::draw(sf::RenderTarget& target, sf::RenderStates state) const
{
	// For now, assume all entities at layer 1
	tilemap.drawLayers(target, state, 0, 1);

	// Draw bullets
	for(auto bulletIter = bullets.begin(); bulletIter != bullets.end(); ++bulletIter) {
		target.draw(*bulletIter, state);
	}

	//Draw all the entities on top of the tiles
	for(unsigned int i = 0; i < activatables.size(); i++) {
		target.draw(*activatables[i], state);
	}
	for(unsigned int i = 0; i < enemies.size(); i++) {
		target.draw(*enemies[i], state);
	}

	//Draw the player
	target.draw(player, state);

	tilemap.drawLayers(target, state, 2, -1);

	//Uncomment to draw attack cones
	/*for(std::list<AttackCone>::const_iterator p = attackCones.begin(); p != attackCones.end(); p++) {
		sf::ConvexShape triangle = p->GetTriangle();
		triangle.setPosition(p->GetOwner().GetPos());
		target.draw(triangle, state);
	}*/
}

bool Level::CheckLevelTransition(std::string& levelName, std::string& spawnName)
{
	if(!isActive) {
		levelName = nextLevel;
		spawnName = nextSpawn;
		return true;
	}
	return false;
}

void Level::SetLevelTransition(const std::string& levelName, const std::string& spawnName)
{
	nextLevel = levelName;
	nextSpawn = spawnName;
	isActive = false;
}
