//Implementation of Player class

#include "Player.h"

#include "Utility/Utility.h"
#include "Level.h"
#include "AttackCone.h"
#include "Bullet.h"
#include "Services/SoundService.h"

const sf::Time Player::attackConeLife = sf::seconds(0.2f);
const sf::Time Player::maxSprintTime = sf::seconds(3.0f);
const float Player::attackConeLength = 50.0f;
const float Player::attackConeSweep = 2*PI/8;
const float Player::regSpeed = 120.0f;
const float Player::sprintSpeed = 200.0f;
const int Player::healthIncrease = 50;
const int Player::baseAttack = 20;
bool Player::soundsRegistered = false;

Player::Player(const sf::Vector2f& pos)
	: Unit("assets/img/Player.png", sf::IntRect(8, 12, 27, 23), sf::IntRect(0, 0, 50, 50))
{
	//When the player starts, he isn't moving anywhere
	SetPos(pos);

	state = CANMOVE;

	isSprinting = false;

	sprintTimer = maxSprintTime;
	souls = 0;

	//Initialize animations (maybe offload some of this data to a text file
	// or something?)
	animManager.AddAnimSet("walk", 1, 16, true);
	animManager.AddAnimSet("attack", 17, 23, false);
}

void Player::RegisterSounds() {
	if(!soundsRegistered) {
		//Initialize sounds
		soundService->AddSound("assets/sound/swoosh.ogg", "swing", false);
		soundService->AddSound("assets/sound/thwack.ogg", "hit", false);
		soundService->AddSound("assets/sound/ow.ogg", "got_hit", false);
		soundsRegistered = true;
	}
}

void Player::Reset()
{
	//When the player starts, he isn't moving anywhere
	dir = sf::Vector2f(0,0);

	isSprinting = false;

	sprintTimer = maxSprintTime;
	health = maxHealth;
	souls = 0;
}

void Player::StartAttack()
{
	if(state != ATTACKING && state != STARTATTACK) {
		state = STARTATTACK;
	}
}

void Player::Attack(Unit& unit)
{
	if(unit.IsAlive()) {
		Unit::Attack(unit);
		soundService->PlaySound("hit");
	}
}

void Player::Shoot(bool shoot) {
	gun.ToggleFiring(shoot);
}

void Player::SetFacing(float angle) {
	facing = angle * TO_DEG;
}

bool Player::IsAlive()
{
	return Unit::IsAlive();
}

void Player::RemoveHealth(int amount)
{
	Unit::RemoveHealth(amount);
	soundService->PlaySound("got_hit");
}

void Player::MoveLeft(bool start)
{
	dir.x -= (start ? 1.0f : -1.0f);
}
void Player::MoveRight(bool start)
{
	dir.x += (start ? 1.0f : -1.0f);
}
void Player::MoveUp(bool start)
{
	dir.y -= (start ? 1.0f : -1.0f);
}
void Player::MoveDown(bool start)
{
	dir.y += (start ? 1.0f : -1.0f);
}

void Player::Sprint(bool start)
{
	isSprinting = start;
}

void Player::AddHealth(int recovery)
{
	health = std::min(maxHealth, health+recovery);
}

void Player::AddEnergy(int recovery)
{
	sf::Time newtime = sf::microseconds((sf::Int64)
		(sprintTimer.asMicroseconds()+recovery/100.0f*maxSprintTime.asMicroseconds()));
	sprintTimer = std::min(maxSprintTime, newtime);
}

void Player::Update(Level& level, const sf::Time& timePassed)
{
	timer += timePassed;
	bool moving = dir.x != 0 || dir.y != 0;
	float moveSpeed;
	sf::Vector2f movement;

	if(isSprinting) {
		moveSpeed = sprintSpeed*timePassed.asSeconds();
		sprintTimer = std::max(sf::Time::Zero, sprintTimer-timePassed);
	} else {
		moveSpeed = regSpeed*timePassed.asSeconds();
	}

	movement.x = dir.x == 0 ? 0 : dir.x * moveSpeed / magnitude(dir);
	movement.y = dir.y == 0 ? 0 : dir.y * moveSpeed / magnitude(dir);

	if(state == CANMOVE && moving) {
		PlayAnim("walk", timePassed);
		move(level, movement);
		sprite.setRotation(TO_DEG * atan2f(movement.y, movement.x));
	} else if(state == ATTACKING) {
		PlayAnim("attack", timePassed);
		if(timer >= attackConeLife) {
			state = CANMOVE;
		}
	} else if(state == STARTATTACK) {
		soundService->PlaySound("swing");

		state = ATTACKING;
		timer = sf::Time::Zero;

		level.AddAttack(AttackCone(attackConeLife, GetPos(), attackConeLength,
			sprite.getRotation()*TO_RAD - attackConeSweep/2.0f,
			sprite.getRotation()*TO_RAD + attackConeSweep/2.0f, *this));

		state = ATTACKING;
	} else {
		//Something's probably wrong
		PlayAnim("idle", timePassed);
	}

	//Regenerate stamina when not running
	if(!isSprinting) {
		sprintTimer = std::min(maxSprintTime, sprintTimer+timePassed);
	}
	if(sprintTimer <= sf::seconds(0.0f)) {
		isSprinting = false;
	}

	//Fire the gun
	gun.Update(level, timePassed, GetPos(), facing);
}
