/**
 * \brief Represents the player's avatar.
 * \author Edward Lu
 * \date January 28 2013
 *
 * This is a subclass of sf::Drawable so that it can be drawn by itself.
 */

#ifndef __PLAYER_H
#define __PLAYER_H

#include <SFML/Graphics.hpp>
#include <set>

#include "Unit.h"
#include "Gun.h"
class Level;
class AttackCone;

class Player : public Unit 
{
//Public attributes; visible to everyone
public:
	/**
	 * Init constructor.
	 * Initializes the player with no graphical component whatsoever - Make sure to
	 *  call LoadTexture() before drawing the player!
	 */
	Player(const sf::Vector2f& pos);

	/**
	 * Starts or stops the player moving left.
	 * \param start If true, the player starts moving left; otherwise, they stop
	 *  moving left.
	 */
	void MoveLeft(bool start);
	/**
	 * Starts or stops the player moving right.
	 * \param start If true, the player starts moving right; otherwise, they stop
	 *  moving right.
	 */
	void MoveRight(bool start);
	/**
	 * Starts or stops the player moving up.
	 * \param start If true, the player starts moving up; otherwise, they stop
	 *  moving up.
	 */
	void MoveUp(bool start);
	/**
	 * Starts or stops the player moving down.
	 * \param start If true, the player starts moving down; otherwise, they stop
	 *  moving down.
	 */
	void MoveDown(bool start);

	/**
	 * Turns the player's sprinting on or off.
	 * \param start True if the player should start sprinting, false otherwise.
	 */
	void Sprint(bool start);

	/**
	 * Gets the player's AttackCone.
	 * \return An AttackCone representing the area the player would hit.
	 */
	void StartAttack();

	/**
	 * Makes the player hit another Movable with his attack.
	 * 
	 * Overriden so we can play some sounds.
	 * \param movable The Movable to attack.
	 */
	virtual void Attack(Unit& unit);

	/**
	 * Sets player shooting flag.
	 **/
	void Shoot(bool start);

	/**
	 * Sets player facing
	 **/
	void SetFacing(float angle);

	/**
	 * Makes the player take damage from another attack.
	 *
	 * Overriden so we can play a sound.
	 * \param amount The amount of health to remove.
	 */
	void RemoveHealth(int amount);

	/**
	 * Updates the player's position.
	 * This function should be called once per frame.
	 */
	virtual void Update(Level& level, const sf::Time& timePassed);

	/**
	 * Adds to the player's health.
	 * \param recovery The amount to add.
	 */
	void AddHealth(int recovery);

	/**
	 * Adds to the player's energy level.
	 * \param recovery The amount to add.
	 */
	void AddEnergy(int energy);

	/**
	 * Resets the player, clearing all souls, keys etc. he has.
	 */
	void Reset();

	/**
	 * Checks if the player is alive/not a monster.
	 */
	virtual bool IsAlive();

//Private attributes; visible only within this class
private:
	virtual void RegisterSounds();

	enum {
		CANMOVE,
		STARTATTACK,
		ATTACKING
	} state;

	/** Current movement direction */
	sf::Vector2f dir;

	/** True when the player is sprinting, false when he is walking at regSpeed */
	bool isSprinting;

	/** Amount of fuel the player has left */
	sf::Time sprintTimer;
	/** Number of pieces of soul the player has. */
	int souls;

	float facing;

	/** Gun */
	Gun gun;

	/** Timer for states */
	sf::Time timer;

	/** Time-to-live for any attack the player initiates. */
	static const sf::Time attackConeLife;
	/** Max burst sprinting time */
	static const sf::Time maxSprintTime;
	/** Length of the player's attack cone */
	static const float attackConeLength;
	/** Angle of the player's attack cone */
	static const float attackConeSweep;
	/** Regular walking speed, in pixels per second. */
	static const float regSpeed;
	/** Running speed, in pixels per second. */
	static const float sprintSpeed;
	/** Amount of health the player gains by picking up a soul. */
	static const int healthIncrease;
	/** Base attack power of the player. */
	static const int baseAttack;

	static bool soundsRegistered;
};

#endif
