
#include "TiledMapCollision.h"

#include <unordered_map>
#include <sstream>
#include <fstream>
#include <cfloat>

#include "Utility/Utility.h"

static sf::Vector2i vecFromJson(Json::Value value) {
	return sf::Vector2i(value["x"].asUInt(), value["y"].asUInt());
}

static Edge<int> edgeFromJson(Json::Value edge, Json::Value corners) {
	sf::Vector2i c1 = vecFromJson(corners[edge["c1"].asUInt()]);
	sf::Vector2i c2 = vecFromJson(corners[edge["c2"].asUInt()]);
	return Edge<int>(c1, c2);
}

static bool check_del(const Edge<int> &e1, const Edge<int> &e2) {
	//Delete an edge if it is the same as another edge
	return (e1.c1 == e2.c2 && e1.c2 == e2.c1) || (e1.c1 == e2.c2 && e1.c2 == e2.c1);
}

static bool check_merge(const Edge<int> &e1, const Edge<int> &e2, Edge<int> &merge) {
	//Merge edges if they are facing the same direction and have one point in common
	bool merged = false;
	if(std::abs(dot(e1.getNormal(), e2.getNormal()) - (magnitude(e1.getNormal()) * magnitude(e2.getNormal()))) <= EPSILON) {
		if((e1.c1 == e2.c1 && e1.c2 == e2.c2) ||
				(e1.c1 == e2.c2 && e1.c2 == e2.c1)) {
			//The edges are the same; handle gracefully
			merge.c1 = e1.c1;
			merge.c2 = e1.c2;
			merged = true;
		//Normal information is encoded in the direction of the edges;
		// if the edges are going in the same directions, they have equal
		// normals
		} else if(e1.c1 == e2.c2) {
			merge.c1 = e2.c1;
			merge.c2 = e1.c2;
			merged = true;
		} else if(e1.c2 == e2.c1) {
			merge.c1 = e1.c1;
			merge.c2 = e2.c2;
			merged = true;
		} 
	}

	return merged;
}

//Takes the name of a tileset annotation file and returns a map of tileset tile IDs to
// edges within that tileset tile.
static std::unordered_map<unsigned, std::vector<Edge<int>>> GetEdgeAnnotation(Json::Value root, sf::Vector2i tileSize) {
	//Iterate through JSON file and convert
	std::unordered_map<unsigned, std::vector<Edge<int>>> edges;
	for(Json::ValueIterator val = root.begin(); val != root.end(); val++) {
        unsigned tileId = 0;
        std::stringstream sstream(val.key().asString());
        sstream >> tileId;
        std::vector<Edge<int>> &tileEdges = edges[tileId];

        //Parse the json
        Json::Value tileJson = *val;
        Json::Value specialJson = tileJson["special"];
        Json::Value edgeJson = tileJson["edges"];
        Json::Value cornerJson = tileJson["points"];

        if(!specialJson.isNull()) {
            //The tile is one of a predefined type
            if(specialJson.asString() == "Full") {
                //Full tile (empty space)
                sf::Vector2i tl(0, 0);
                sf::Vector2i tr(tileSize.x, 0);
                sf::Vector2i br(tileSize);
                sf::Vector2i bl(0, tileSize.y);
                tileEdges.push_back(Edge<int>(tl, tr));
                tileEdges.push_back(Edge<int>(tr, br));
                tileEdges.push_back(Edge<int>(br, bl));
                tileEdges.push_back(Edge<int>(bl, tl));
            }
        } else if(!edgeJson.isNull() && !cornerJson.isNull() &&
                edgeJson.isArray() && cornerJson.isArray()) {
            // Parse the edges of the open space
            for(unsigned i = 0; i < edgeJson.size(); i++) {
                tileEdges.push_back(edgeFromJson(edgeJson[i], cornerJson));
            }
        }
    }
	return edges;
}

//Returns a mapping of layer tile ids to tileset tile ids.
static std::vector<unsigned> MapTiles(const Tmx::TileLayer *collisionLayer) {
	sf::Vector2i layerSize(collisionLayer->GetWidth(), collisionLayer->GetHeight());
    std::vector<unsigned> tileMap(layerSize.x * layerSize.y);
    for(unsigned y = 0; y < layerSize.y; y++) {
        for(unsigned x = 0; x < layerSize.x; x++) {
            // Tile id within the tileset
            unsigned tileId = collisionLayer->GetTile(x, y).id;
            // Tile id within the layer
            unsigned layerTileId = y * layerSize.x + x;
            tileMap[layerTileId] = tileId;
        }
    }
    return tileMap;
}

//Simplifies (merges, deletes) edges from tiles into edges that can be turned into polygons.
static std::vector<Edge<int>> Simplify(
        const std::unordered_map<unsigned, std::vector<Edge<int>>> &tileToEdgeMap,
        const std::vector<unsigned> &tileToTileMap,
        sf::Vector2i layerSize, sf::Vector2i tileSize) {
	std::vector<Edge<int>> edges;
    std::vector<std::vector<bool>> tileToEdgeDelMap(tileToTileMap.size());

	//Iterate over all tiles with marked edges
    for(unsigned layerTileId = 0; layerTileId < tileToTileMap.size(); layerTileId++) {
		unsigned tileId = tileToTileMap[layerTileId];
		
		//Extract the tile coordinates from the tile id
		unsigned y = layerTileId / layerSize.x;
		unsigned x = layerTileId % layerSize.x;

        auto iter = tileToEdgeMap.find(tileId);
        if(iter == tileToEdgeMap.end()) {
            // Not a problem - just means this tile has no open space
            continue;
        }

		const std::vector<Edge<int>> &edges_1 = iter->second;

		//Iterate over the open space polygon of the current tile
		for(unsigned i = 0; i < edges_1.size(); i++) {
			Edge<int> mergeCandidate;
			bool merge = false;
			bool del = false;
			Edge<int> e1 = edges_1[i].offset(sf::Vector2i(x * tileSize.x, y * tileSize.y));

			if(tileToEdgeDelMap[layerTileId].size() > 0) {
                if(tileToEdgeDelMap[layerTileId][i]) {
                    //Edge has already been deleted before, skip it
                    continue;
                }
			} else {
                tileToEdgeDelMap[layerTileId].resize(edges_1.size(), false);
            }

			//Only edges in tiles adjacent to the current tile
			// can be merged or deleted with edges of the current tile
			for(int xoff = -1; xoff <= 1 && !del; xoff++) {
				for(int yoff = -1; yoff <= 1 && !del; yoff++) {
					unsigned layerTileIdOff = (y + yoff) * layerSize.x + (x + xoff);

                    //Skip if current tile or beyond bounds
					if(layerTileIdOff == layerTileId ||
                            layerTileIdOff >= tileToTileMap.size()) {
                        continue;
                    }

                    unsigned tileIdOff = tileToTileMap[layerTileIdOff];
                    auto iter = tileToEdgeMap.find(tileIdOff);
                    // Skip if tile has no open space (no edges)
                    if(iter == tileToEdgeMap.end()) {
                        continue;
                    }

					const std::vector<Edge<int>> &edges_2 = iter->second;
					for(unsigned j = 0; j < edges_2.size() && !del; j++) {
                        Edge<int> e2 = edges_2[j].offset(
                                sf::Vector2i((x+xoff) * tileSize.x, (y+yoff) * tileSize.y));

                        if(tileToEdgeDelMap[layerTileIdOff].size() > 0) {
                            if(tileToEdgeDelMap[layerTileIdOff][j]) {
                                //Edge has already been deleted before, skip it
                                continue;
                            }
                        } else {
                            tileToEdgeDelMap[layerTileIdOff].resize(edges_2.size(), false);
                        }

						if(check_del(e1, e2)) {
							//The edges are equal and normals opposite; delete both
							// and continue on (check the previous for loop conditions)
							tileToEdgeDelMap[layerTileId][i] = true;
							tileToEdgeDelMap[layerTileIdOff][j] = true;
							del = true;
						} 
					}
				}
			}

			//The edge was not deleted - try merging it
			if(!del) {
				// Check for merge with previous edges
				for(unsigned j = 0; j < edges.size() && !del && !merge; j++) {
					Edge<int> &e2 = edges[j];
					if(check_merge(e1, e2, mergeCandidate)) {
                        tileToEdgeDelMap[layerTileId][i] = true;
						//Instead of deleting/inserting, just replace
						// the edge that we merged with
						edges[j] = mergeCandidate;
						merge = true;
					}
				}

				if(!merge) {
					//Nothing; insert the edge as-is. It might be merged later,
					// but we are sure it won't be deleted.
					edges.push_back(e1);
                    tileToEdgeDelMap[layerTileId][i] = true;
				}
			}
		}
	}

	return edges;
}

TiledMapCollision::TiledMapCollision() {}
TiledMapCollision::TiledMapCollision(Tmx::Map &map) {
    LoadMap(map);
}

void TiledMapCollision::LoadPolylineGroup(const Tmx::ObjectGroup *objectGroup) {
    //Iterate through all objects in the object group
    for (int j = 0; j < objectGroup->GetNumObjects(); ++j) {
        //Grab objects from the group
        const Tmx::Object *object = objectGroup->GetObject(j);
        const Tmx::Polyline *polyline = object->GetPolyline();

        if(!polyline) {
            fprintf(stderr, "WARNING: Non-polyline object found in collision layer, ignoring...");
        } else {
            sf::Vector2i origin(object->GetX(), object->GetY());
            sf::Vector2i prevPoint(polyline->GetPoint(0).x, polyline->GetPoint(0).y);
            for(int k = 1; k < polyline->GetNumPoints(); k++) {
                sf::Vector2i point(polyline->GetPoint(k).x, polyline->GetPoint(k).y);
                edges.emplace_back(origin + prevPoint, origin + point);
                point = prevPoint;
            }
        }
    }
}

void TiledMapCollision::LoadMap(Tmx::Map &map) {
	const Tmx::TileLayer *collisionLayer = NULL;
	//Figure out which layer is collision
	for(unsigned i = 0; i < map.GetNumLayers(); i++) {
		const Tmx::TileLayer *layer = map.GetTileLayer(i);
		if(layer->GetProperties().HasProperty("Collision") &&
				layer->GetProperties().GetIntProperty("Collision")) {
			collisionLayer = layer;
			break;
		}
	}

    bool good = false;

	if(collisionLayer) {
        if(!collisionLayer->GetProperties().HasProperty("TileDef")){
            fprintf(stderr, "Collision layer %s doesn't have a TileDef property!\n",
                    collisionLayer->GetName().c_str());
        }

        std::string annotationFilename = collisionLayer->GetProperties().GetStringProperty("TileDef");
        annotationFilename = "assets/" + annotationFilename;
        sf::Vector2i layerSize(collisionLayer->GetWidth(), collisionLayer->GetHeight());
        sf::Vector2i tileSize(map.GetTileWidth(), map.GetTileHeight());

        //Grab the json file
        std::ifstream jsonfile(annotationFilename);
        if(!jsonfile.good()) {
            fprintf(stderr, "Failed to read annotation file %s\n", annotationFilename.c_str());
        }

        //Parse the file
        Json::Value root;
        Json::Reader reader;
        bool parsed = reader.parse(jsonfile, root);
        if(!parsed) {
            fprintf(stderr, "Failed to parse annotation file %s: %s\n",
                    annotationFilename.c_str(), reader.getFormattedErrorMessages().c_str());
        }

        std::unordered_map<unsigned, std::vector<Edge<int>>> tileToEdgeMap =
            GetEdgeAnnotation(root, tileSize);
        std::vector<unsigned> tileToTileMap = MapTiles(collisionLayer);

        this->edges = Simplify(tileToEdgeMap, tileToTileMap, layerSize, tileSize);

        good = true;
    }
    
    if(!good) {
        bool collision = false;
        //Maybe it's an object layer instead
        for (int i = 0; i < map.GetNumObjectGroups() && !collision; ++i) {
            const Tmx::ObjectGroup *objectGroup = map.GetObjectGroup(i);
            collision = objectGroup->GetProperties().GetIntProperty("Collision");

            if(collision) {
                sf::Vector2i mapSize(map.GetWidth(), map.GetWidth());
                sf::Vector2i tileSize(map.GetTileWidth(), map.GetTileHeight());
                LoadPolylineGroup(objectGroup);

                //Add map bounds to edges, just in case
                sf::Vector2i tl(0.0f, 0.0f);
                sf::Vector2i tr(0.0f, mapSize.y*tileSize.y);
                sf::Vector2i br(mapSize.x*tileSize.x, mapSize.y*tileSize.y);
                sf::Vector2i bl(mapSize.x*tileSize.x, 0.0f);
                edges.emplace_back(tl, tr);
                edges.emplace_back(tr, br);
                edges.emplace_back(br, bl);
                edges.emplace_back(bl, tl);

                good = true;
            }
        }
	}

    if(!good) {
        fprintf(stderr, "Failed to find a collision layer when loading tiled map\n");
    }
}

bool TiledMapCollision::GetCollide(sf::Vector2f p1, sf::Vector2f p2, sf::Vector2f& closestIntersect) const {
	float closestMag = FLT_MAX;
	sf::Vector2f dir = p2 - p1;
	for(std::size_t j = 0; j < edges.size(); j++) {
		Edge<int> edge = edges[j];
		sf::Vector2f c1(edge.c1.x, edge.c1.y);
		sf::Vector2f c2(edge.c2.x, edge.c2.y);
		sf::Vector2f intersection;

        bool intersected =  lineIntersect(p1, p2, c1, c2, intersection);
		float intersectMag = magnitude(p1 - intersection);

		/* If we intersected on the segment and this is a closer intersection
		 * than any prior, take this as the new closest intersection. Some
		 * weird corner cases here: If the ray is exactly horizontal or
		 * vertical, we've got to check if the intersection is close enough
		 * (floating point inaccuracy will make the sign be off by a little
		 * bit). If the ray is shot from a point on the line, the sign will be
		 * 0. */
		if(intersected &&
                intersectMag < closestMag &&
				(std::abs(p2.x - p1.x) < EPSILON ||
				 sign(intersection.x - p1.x) == sign(dir.x)) &&
				(std::abs(p2.y - p1.y) < EPSILON ||
				 sign(intersection.y - p1.y) == sign(dir.y)) &&
				onLine(c1, c2, intersection, EPSILON)) {
			closestIntersect = intersection;
			closestMag = intersectMag;
		}
	}

	return closestMag != FLT_MAX;
}

const std::vector<Edge<int>> &TiledMapCollision::getEdges() {
    return edges;
}
