
#include "Tilemap.h"

std::unordered_map<std::string, sf::Texture> Tilemap::tilesetTextureCache;


sf::Texture &Tilemap::getTilesetTexture(const std::string &loc) {
	auto cacheIter = tilesetTextureCache.find(loc);
	if(cacheIter != tilesetTextureCache.end()) {
		return cacheIter->second;
	} else {
		tilesetTextureCache[loc] = sf::Texture();
		sf::Texture &texture = tilesetTextureCache[loc];

		if(!texture.loadFromFile(loc)) {
			fprintf(stderr, "Error loading tileset from %s", loc.c_str());
			// argh
			throw 1;
		}

		return texture;
	}
}

Tilemap::Tilemap() { }

Tilemap::Tilemap(Tmx::Map &map) {
	LoadMap(map);
}

void Tilemap::LoadMap(Tmx::Map &map) {
	//Grab all the tileset textures
	std::vector<sf::Texture*> tilesetTextures(map.GetNumTilesets());
	sf::Vector2u mapSize(map.GetWidth(), map.GetHeight());
	sf::Vector2u tileSize(map.GetTileWidth(), map.GetTileHeight());
	for(int i = 0; i < map.GetNumTilesets(); i++) {
		const Tmx::Tileset *tileset = map.GetTileset(i);
		sf::Texture &texture = getTilesetTexture(BASEMAPDIR + tileset->GetImage()->GetSource());
		tilesetTextures[i] = &texture;
	}

	//Draw to the tilemap texture
	for(int i = 0; i < map.GetNumTileLayers(); i++)
	{
		sf::Sprite tilesetSprite;
		const Tmx::TileLayer *layer = map.GetTileLayer(i);
		int layerNum = 0;
		if(layer->GetProperties().HasProperty("Layer")) {
			layerNum = layer->GetProperties().GetIntProperty("Layer");
		} else {
			fprintf(stderr, "WARNING: Layer \'%s\' has no layer number, setting to 0\n",
					layer->GetName().c_str());
		}

		if(tilemapTextures.find(layerNum) == tilemapTextures.end()) {
			tilemapTextures[layerNum].create(mapSize.x*tileSize.x, mapSize.y*tileSize.y);
		}

		//Render tiles of the layer
		for(int y = 0; y < layer->GetHeight(); y++)
		{
			for(int x = 0; x < layer->GetWidth(); x++)
			{
				//If there is no tile on this location in the map, GetTileTilesetIndex()
				// returns -1, so just do nothing
				Tmx::MapTile tile = layer->GetTile(x, y);
				//printf("tilesetIdx: %d\n", tilesetIdx);
				if(tile.tilesetId >= 0)
				{
					//Map the tile index to a coordinate in the tileset
					const Tmx::Tileset* tileset = map.GetTileset(tile.tilesetId);
					unsigned int tileIdx = tile.id;
					int tilesetWidth = tileset->GetImage()->GetWidth() / tileSize.x;

					sf::Vector2u locTile(tileIdx % tilesetWidth * tileSize.x, tileIdx / tilesetWidth * tileSize.y);

					tilesetSprite.setTexture(*tilesetTextures[tile.tilesetId]);

					//Get the part of the sprite corresponding to the tile
					tilesetSprite.setTextureRect(sf::IntRect(locTile.x, locTile.y, map.GetTileWidth(), map.GetTileHeight()));
					//Set the tile's position to the correct place
					tilesetSprite.setPosition((float)x*map.GetTileWidth(), (float)y*map.GetTileHeight());

					//Draw that tile
					tilemapTextures[layerNum].draw(tilesetSprite);
				}
			}
		}
	}

	//We're finished drawing to them - let SFML finalize the textures
	for(auto textureIter = tilemapTextures.begin(); textureIter != tilemapTextures.end(); textureIter++) {
		textureIter->second.display();
	}
}

void Tilemap::drawLayers(sf::RenderTarget& target, sf::RenderStates state, unsigned from, unsigned to) const {
	auto iter = tilemapTextures.begin();
	while(iter != tilemapTextures.end() && iter->first <= to) {
		if(iter->first >= from) {
			target.draw(sf::Sprite(iter->second.getTexture()), state);
		}
		++iter;
	}
}
