
#include "Navmesh.h"
#include "Utility/Utility.h"

#include "polypartition/polypartition.h"

#include <cfloat>
#include <fstream>
#include <sstream>
#include <unordered_map>

//Converts the polygons from Simplify (or any vector of Edge<int>s, for that matter)
// into a list of TPPLPoly objects.
static std::list<TPPLPoly> ConvertToPoly(const std::vector<Edge<int>> &edges) {
	std::vector<bool> processed(edges.size(), false);
	std::list<TPPLPoly> polygons;

	for(unsigned i = 0; i < edges.size(); i++) {
		std::vector<sf::Vector2i> points;
		Edge<int> first = edges[i];
		Edge<int> e1 = first;
		bool closed = false;

		if(processed[i]) { continue; }

		bool found = false;
		do {
			found = false;
			//Traverse edges
			for(unsigned j = 0; !found && j < edges.size(); j++) {
				Edge<int> e2 = edges[j];

				//Edge can't be part of two polys
				if(processed[j]) { continue; }

				if(e1.c2 == e2.c1) {
					points.push_back(e1.c2);
					//Found next edge
					e1 = e2;
					processed[j] = true;
					found = true;

					if(e2.c1 == first.c1) {
						//Back to first edge
						closed = true;
					}
				}
			}

			//If we didn't find a connecting edge, something wrong
			if(!found) {
				fprintf(stderr, "WARNING: Found unclosed poly when trying to partition polys");
			}
		} while(!closed && found);

		if(closed) {
			TPPLPoly polygon;
			polygon.Init(points.size());
			for(unsigned j = 0; j < points.size(); j++) {
				polygon[j] = TPPLPoint(points[j].x, points[j].y);
			}

			if(polygon.GetOrientation() == TPPL_CW) {
				polygon.SetHole(true);
			} else {
				polygon.SetHole(false);
			}

			polygons.push_back(polygon);
		}
	}

	return polygons;
}

//Partitions a list of TPPLPoly objects into polygons.
static std::vector<std::vector<sf::Vector2f>> Partition(std::list<TPPLPoly> in_polys) {
	TPPLPartition partition;
	std::list<TPPLPoly> out_polys;

	//if(!partition.Triangulate_EC(&polys, &triangles)) {
	if(!partition.ConvexPartition_HM(&in_polys, &out_polys)) {
		fprintf(stderr, "WARNING: Couldn't partition polygons\n");
	}

	//Convert the TPPLPoly list to vector of polygons
	std::vector<std::vector<sf::Vector2f>> conv_polys;
	for(auto iter = out_polys.begin(); iter != out_polys.end(); iter++) {
		TPPLPoly &poly = *iter;
		std::vector<sf::Vector2f> points;
		for(int i = 0; i < poly.GetNumPoints(); i++) {
			points.push_back(sf::Vector2f(poly[i].x, poly[i].y));
		}
		conv_polys.push_back(points);
	}

	return conv_polys;
}

Navmesh::Navmesh() {
}

void Navmesh::LoadEdges(const std::vector<Edge<int>> &edges) {
	std::list<TPPLPoly> mapPolys = ConvertToPoly(edges);
	mapEdges(mapPolys);
	polys = Partition(mapPolys);
	adjacencyList = makeGraph(polys, edges);
}

//Makes a graph out of a vector of polygons. Polygons are points; any polygons that share
// a physical edge are connected by a graph edge.
//This is the naive implementation. Might get slow for large sets of polygons.
std::vector<std::vector<Navmesh::AdjNode>> Navmesh::makeGraph(const std::vector<std::vector<sf::Vector2f>> &polys, const std::vector<Edge<int>> &edges) {
	std::vector<std::vector<Navmesh::AdjNode>> adjacencyList(polys.size());

	for(unsigned i = 0; i < polys.size(); i++) {
		for(unsigned j = 0; j < polys[i].size(); j++) {
			sf::Vector2f c1 = polys[i][j];
			sf::Vector2f c2 = polys[i][(j+1)%polys[i].size()];
			bool done = false;

			for(unsigned k = 0; k < edges.size() && !done; k++) {
				if((close(edges[k].c1, c1) && close(edges[k].c2, c2)) ||
						(close(edges[k].c2, c1) && close(edges[k].c1, c2))) {
					//We can skip this edge, since it's not matched up with another poly
					done = true;
				}
			}

			if(done) { continue; }

			//Otherwise, compare to the rest of the polys
			for(unsigned k = 0; k < polys.size() && !done; k++) {
				//Skip the current poly
				if(k == i) { continue; }
				for(unsigned l = 0; l < polys[k].size() && !done; l++) {
					sf::Vector2f c3 = polys[k][l];
					sf::Vector2f c4 = polys[k][(l+1)%polys[k].size()];
					if((close(c1, c3) && close(c2, c4)) || (close(c2, c3) && close(c1, c4))) {
						Navmesh::AdjNode node(k, l, j);
						adjacencyList[i].push_back(node);
						done = true;
					}
				}
			}

			if(!done) {
				fprintf(stderr, "WARNING: didn't find a corresponding portal for %u,%u\n", i, j);
			}
		}
	}
	return adjacencyList;
}

void Navmesh::mapEdges(const std::list<TPPLPoly> &polys) {
	for(auto p = polys.begin(); p != polys.end(); p++) {
		const TPPLPoly &poly = *p;
		unsigned polysize = poly.GetNumPoints();
		for(unsigned i = 0; i < polysize; i++) {
			TPPLPoint p1 = poly[(polysize + i - 1)%polysize];
			TPPLPoint p2 = poly[(i + 1)%polysize];
			std::pair<sf::Vector2f, sf::Vector2f> pair(sf::Vector2f(p1.x, p1.y), sf::Vector2f(p2.x, p2.y));
			sf::Vector2f point(poly[i].x, poly[i].y);
			edgeFinderMap[point] = pair;
		}
	}
}

//Returns a g_value, for use in the TA* algorithm. See Demyen.
float Navmesh::g_val(sf::Vector2f start, sf::Vector2f end,
		CostPair p_cost, float h_val, AdjNode parent, AdjNode node,
		bool trust_parent) {
	const std::vector<sf::Vector2f> &poly = polys[node.poly];
	const std::vector<sf::Vector2f> &ppoly = polys[parent.poly];
	sf::Vector2f p1 = poly[node.portal];
	sf::Vector2f p2 = poly[(node.portal+1)%poly.size()];
	sf::Vector2f pp1 = ppoly[parent.portal];
	sf::Vector2f pp2 = ppoly[(parent.portal+1)%ppoly.size()];

	float g1 = lineDistance(start, p1, p2);
	float g2 = magnitude(start - end) - h_val;
	float g3 = p_cost.first + p_cost.second - h_val;
	float g4 = FLT_MIN;
	if(trust_parent) {
		g4 = p_cost.first + lineDistance(p1, p2, pp1, pp2);
	}

	return std::max(g1, std::max(g2, std::max(g3, g4)));
}

//Returns an h-value, for use in the TA* algorithm. See Demyen.
float Navmesh::h_val(sf::Vector2f goal, AdjNode node) {
	std::vector<sf::Vector2f> &poly = polys[node.poly];
	sf::Vector2f p1 = poly[node.portal];
	sf::Vector2f p2 = poly[(node.portal+1)%poly.size()];
	return lineDistance(goal, p1, p2);
}

//Runs TA* on the polygon graph. See Demyen.
//Most of this is normal A* on any old graph, except for the part where we
// measure distance traveled (h_value, g_value).
std::vector<Navmesh::PathNode> Navmesh::findPolyPath(sf::Vector2f from, sf::Vector2f to, float radius) {
	prioq_t open_set;
	unsigned start = 0, end = 0;
	bool found_start = false, found_end = false;
	std::unordered_map<unsigned, CostPair> cur_cost;
	std::unordered_map<unsigned, AdjNode> came_from;
	std::vector<PathNode> list;

	// Find the polygon the start point is in
	for(unsigned i = 0; i < polys.size(); i++) {
		if(contains(polys[i], from)) {
			start = i;
			found_start = true;
		}
		if(contains(polys[i], to)) {
			end = i;
			found_end = true;
		}
	}

	//If we didn't find a start or an end, return empty list
	if(!found_start || !found_end) {
		return list;
	}

	//If the start element is the end element, just return a path
	// of one polygon
	if(start == end) {
		list.push_back(Navmesh::PathNode(start, 0, 0));
		return list;
	}

	//Initialize
	AdjNode startNode(start, 0, 0);
	came_from[start] = startNode;
	cur_cost[start] = CostPair(0.0f, h_val(to, startNode));
	open_set.push(std::pair<float, AdjNode>(0.0f, startNode));

	//While we still have nodes to explore...
	while(!open_set.empty()) {
		//Grab top of open set
		std::pair<float, AdjNode> val = open_set.top();
		//"pnode" for previous node
		AdjNode pnode = val.second;
		CostPair pcost = cur_cost[pnode.poly];
		std::vector<AdjNode> &adjacent = adjacencyList[pnode.poly];
		open_set.pop();

		for(unsigned i = 0; i < adjacent.size(); i++) {
			AdjNode node = adjacent[i];
			//Get cost of traveling from this node to end (heuristic)
			float hval = h_val(to, node);
			//Get cost of traveling from start to this node
			float gval = g_val(from, to, pcost, hval, pnode, node, pnode.poly != start);
			//Priority represents estimated distance from start to end
			// using path through this node
			float prio = gval + hval;

			auto cost_iter = cur_cost.find(node.poly);

			std::vector<sf::Vector2f> &poly = polys[node.poly];
			sf::Vector2f p1 = poly[node.portal];
			sf::Vector2f p2 = poly[(node.portal+1)%poly.size()];
			float portal_size = magnitude(p2 - p1);

			if(radius*2 < portal_size && (cost_iter == cur_cost.end() ||
					prio < cost_iter->second.first + cost_iter->second.second)) {
				//Portal is big enough, plus
				//Haven't gone through adjacent node yet, or it's better to go
				// through this node than whatever node we went through before
				cur_cost[node.poly] = CostPair(gval, hval);

				//XXX: Debug? 
				prios[node.poly] = std::pair<float, float>(gval, hval);

				open_set.push(std::pair<float, AdjNode>(prio, node));
				//We came from the previous node, using this node's portal
				came_from[node.poly] = AdjNode(pnode.poly, node.adj_portal, node.portal);
			}
		}
	}

	if(came_from.find(end) == came_from.end()) {
		//We never got to the end poly; no path
		return list;
	}

	//Reconstitute the polygon path by retracing our steps and 
	// swapping the portals back to the right orders
	AdjNode from_node(end, 0, 0);
	AdjNode cur_node = from_node;
	do {
		cur_node = from_node;
		from_node = came_from[cur_node.poly];

		list.push_back(PathNode(cur_node.poly, from_node.adj_portal, cur_node.portal));
	} while(cur_node.poly != from_node.poly);

	std::reverse(list.begin(), list.end());
	return list;
}

sf::Vector2f Navmesh::pushAway(sf::Vector2f point, float distance) {
	//Push the point away from the corner by taking average of normals
	// We don't have explicit normals , but we know that these corners
	// must be part of holes, so the normals are rotated by 90 degrees
	// That's equivalent to (a, b) -> (-b, a)
	auto iter = edgeFinderMap.find(point);
	if (iter == edgeFinderMap.end()) {
		fprintf(stderr, "WARNING: Couldn't find corner to push away from at (%g, %g)\n", point.x, point.y);
		return point;
	} else {
		auto pair = iter->second;
		sf::Vector2f e1 = point - pair.first;
		sf::Vector2f e2 = pair.second - point;
		sf::Vector2f n1 = normalize(sf::Vector2f(-e1.y, e1.x));
		sf::Vector2f n2 = normalize(sf::Vector2f(-e2.y, e2.x));
		sf::Vector2f pushDir = normalize(n1 + n2);
		sf::Vector2f newPoint = point + pushDir * distance;
		return newPoint;
	}
}

sf::Vector2f Navmesh::getNextLeft(PathNode node, float radius) {
	sf::Vector2f point = polys[node.poly][node.next_portal];
	return pushAway(point, radius);
}

sf::Vector2f Navmesh::getNextRight(PathNode node, float radius) {
	sf::Vector2f point = polys[node.poly][(node.next_portal+1)%polys[node.poly].size()];
	return pushAway(point, radius);
}

std::vector<sf::Vector2f> Navmesh::findPath(sf::Vector2f start, sf::Vector2f end, float radius) {
    return getPath(start, end, findPolyPath(start, end, radius), radius);
}

std::vector<sf::Vector2f> Navmesh::getPath(sf::Vector2f start, sf::Vector2f end, std::vector<Navmesh::PathNode> polypath, float radius) {
	std::vector<sf::Vector2f> path;

	if(polypath.size() <= 1) {
		//No path, signify by empty vector
		return path;
	}

	path.push_back(start);

	bool advanceLeft = true;
	std::vector<sf::Vector2f> poly = polys[polypath[0].poly];
	//"left" is the first vector in the portal going clockwise; "right" is counterclockwise
	sf::Vector2f left_vec = getNextLeft(polypath[0], radius);
	sf::Vector2f right_vec = getNextRight(polypath[0], radius);
	sf::Vector2f prev_left = left_vec;
	sf::Vector2f prev_right = right_vec;
	sf::Vector2f last_node = start;
	float prevAngle = angleBetween(left_vec, right_vec, last_node);
	//Next polygon to try on the l and r sides, updated every loop
	// to prevent stalls
	unsigned lpoly = 1;
	unsigned rpoly = 1;
	//Adjacent polygon on the l and r sides, only updated when the 
	// l and r vectors are updated
	unsigned cur_lpoly = 1;
	unsigned cur_rpoly = 1;

	while(lpoly < polypath.size() || rpoly < polypath.size()) {
		bool checkConditions = true;
		if(advanceLeft) {
			if(lpoly < polypath.size() - 1) {
				//Advance left vector to next polygon
				PathNode node = polypath[lpoly];
				prev_left = left_vec;
				left_vec = getNextLeft(node, radius);
				lpoly++;
			}
			else if(lpoly == polypath.size() - 1) {
				//Treat the last point as a portal, or we miss some cases
				prev_left = left_vec;
				left_vec = end;
				lpoly++;
			} else {
				//We're at the end
				checkConditions = false;
			}
		} else {
			if(rpoly < polypath.size() - 1) {
				//Advance right in the clockwise direction
				PathNode node = polypath[rpoly];
				prev_right = right_vec;
				right_vec = getNextRight(node, radius);
				rpoly++;
			} else if(rpoly == polypath.size() - 1) {
				//Treat the last point as a portal, or we miss some cases
				prev_right = right_vec;
				right_vec = end;
				rpoly++;
			} else {
				//We're at the end
				checkConditions = false;
			}
		}

		if(checkConditions) {
			//Check funnel conditions
			float angle = angleBetween(left_vec, right_vec, last_node);
			if(angle > prevAngle) {
				//Funnel grew, undo
				if(advanceLeft) {
					left_vec = prev_left;
				} else {
					right_vec = prev_right;
				}
			} else if(angle < 0) {
				bool take_left = !advanceLeft;
				//Funnel crossed, new path node; reset
				// Also, grab the edges forming the corner that we hit
				if(take_left) {
					last_node = left_vec;
					lpoly = rpoly = cur_lpoly;
				} else {
					last_node = right_vec;
					rpoly = lpoly = cur_rpoly;
				}

				path.push_back(last_node);
				
				//While loop takes care of an edge case; keep advancing until we 
				// get a non-close node or die
				while(lpoly < polypath.size() - 1 &&
						(close(left_vec, last_node) || close(right_vec, last_node))) {
					PathNode node = polypath[lpoly];
					left_vec = getNextLeft(node, radius);
					right_vec = getNextRight(node, radius);
					prev_left = left_vec;
					prev_right = right_vec;
					prevAngle = angleBetween(left_vec, right_vec, last_node);
					cur_lpoly = cur_rpoly = rpoly = ++lpoly;
					advanceLeft = false;
				}

			    if(lpoly == polypath.size()) {
					//we're done
					rpoly = ++lpoly;
				}
			} else {
				//Shrink funnel
				prevAngle = angle;
				if(advanceLeft) {
					cur_lpoly = lpoly;
				} else {
					cur_rpoly = rpoly;
				}
			}
		}

		advanceLeft = !advanceLeft;
	}

	path.push_back(end);

	return path;
}
