
#include "NavmeshExposer.h"
#include "game/Tiled/Tilemap.h"
#include "game/Tiled/TiledMapCollision.h"
#include "Utility/Utility.h"

#include <sstream>

#include <SFML/Graphics.hpp>
#include <SFML/Window.hpp>

static const float navAgentSize = 10.0f;

int main(int argc, char *argv[]) {
	if(argc < 2) {
		fprintf(stderr, "Enter a tmx file as the argument\n");
		exit(EXIT_FAILURE);
	}

	Tmx::Map map;
	Tilemap tilemap;
    TiledMapCollision mapCollider;
	NavmeshExposer navmesh;
	bool drawLines = true;
	bool drawTris = false;
	bool drawConnections = true;
	sf::Vector2f from, to;
	bool fromSet = false;
	sf::Font font;

	map.ParseFile(argv[1]);

	tilemap.LoadMap(map);
    mapCollider.LoadMap(map);
	navmesh.LoadEdges(mapCollider.getEdges());

	font.loadFromFile("assets/font/LiberationMono-Regular.ttf");

	const std::vector<Edge<int>> &edges = navmesh.getEdges();
	sf::VertexArray lines(sf::Lines, edges.size()*2);

	for(unsigned i = 0; i < edges.size(); i++) {
		sf::Color col(rand()%255, rand()%255, 0);
		sf::Vector2f c1(edges[i].c1.x, edges[i].c1.y);
		sf::Vector2f c2(edges[i].c2.x, edges[i].c2.y);
		sf::Vertex v1(c1, col);
		sf::Vertex v2(c2, col);
		lines[i*2] = v1;
		lines[i*2+1] = v2;
	}

	const std::vector<std::vector<sf::Vector2f>> &polys = navmesh.getPolys();
	std::vector<sf::ConvexShape> shapes;

	for(unsigned i = 0; i < polys.size(); i++) {
		sf::Color col(rand()%255, rand()%255, 0);
		const std::vector<sf::Vector2f> &poly = polys[i];
		sf::ConvexShape shape(poly.size());
		for(unsigned j = 0; j < poly.size(); j++) {
			shape.setPoint(j, poly[j]);
			shape.setFillColor(sf::Color(rand()%255, rand()%255, rand()%255));
		}
		shapes.push_back(shape);
	}


	const std::vector<std::vector<NavmeshExposer::AdjNode>> &adjacencyList = navmesh.getAdjList();
	sf::VertexArray connections(sf::Lines);
	sf::VertexArray path(sf::Lines);
	std::vector<sf::Text> prios;

	for(unsigned i = 0; i < adjacencyList.size(); i++) {
		const std::vector<sf::Vector2f> &poly = polys[i];
		sf::Vector2f shapeMidpoint(0,0);
		for(int k = 0; k < poly.size(); k++) {
			shapeMidpoint.x += (poly[k].x - shapeMidpoint.x) / (k+1);
			shapeMidpoint.y += (poly[k].y - shapeMidpoint.y) / (k+1);
		}
		sf::Color shapeColor = shapes[i].getFillColor();
		sf::Color color = sf::Color(255 - shapeColor.r, 255 - shapeColor.g, 255 - shapeColor.b);

		for(unsigned j = 0; j < adjacencyList[i].size(); j++) {
			const NavmeshExposer::AdjNode &node = adjacencyList[i][j];
			const std::vector<sf::Vector2f> &npoly = polys[node.poly];
			sf::Vector2f p1 = npoly[node.portal];
			sf::Vector2f p2 = npoly[(node.portal+1)%npoly.size()];
			sf::Vector2f lineMidpoint = sf::Vector2f((p1.x + p2.x)/2.0f, (p1.y + p2.y)/2.0f);
			sf::Vertex v1(shapeMidpoint,color);
			sf::Vertex v2(lineMidpoint,color);
			connections.append(v1);
			connections.append(v2);
		}
	}

	sf::RenderWindow window(sf::VideoMode(800, 600), "navmesh test");
	bool up,down,left,right;
	bool running = true;
	sf::View view = window.getDefaultView();
	up = down = left = right = false;
	sf::Clock clock = sf::Clock();

	while(running) {
		sf::Time deltaTime = clock.getElapsedTime();
		clock.restart();

		sf::Event anEvent;
		sf::Vector2f cameraMove(0.0f,0.0f);
		
		while(window.pollEvent(anEvent)) {
			if(anEvent.type == sf::Event::Closed) {
				running = false;
			} else if(anEvent.type == sf::Event::KeyPressed) {
				switch(anEvent.key.code) {
					case sf::Keyboard::W :
						up = true;
						break;
					case sf::Keyboard::S :
						down = true;
						break;
					case sf::Keyboard::A :
						left = true;
						break;
					case sf::Keyboard::D :
						right = true;
						break;
					case sf::Keyboard::Q :
						drawLines = !drawLines;
						break;
					case sf::Keyboard::E :
						drawTris = !drawTris;
						break;
					default:
						break;
				}
			} else if(anEvent.type == sf::Event::KeyReleased) {
				switch(anEvent.key.code) {
					case sf::Keyboard::W :
						up = false;
						break;
					case sf::Keyboard::S :
						down = false;
						break;
					case sf::Keyboard::A :
						left = false;
						break;
					case sf::Keyboard::D :
						right = false;
						break;
					default:
						break;
				}
			} else if(anEvent.type == sf::Event::MouseButtonPressed) {
				if(anEvent.mouseButton.button == sf::Mouse::Left) {
					if(!fromSet) {
						from = window.mapPixelToCoords(sf::Vector2i(anEvent.mouseButton.x, anEvent.mouseButton.y));
						fromSet = true;
					} else {
						to = window.mapPixelToCoords(sf::Vector2i(anEvent.mouseButton.x, anEvent.mouseButton.y));
						std::vector<NavmeshExposer::PathNode> polypath = navmesh.findPolyPath(from, to, navAgentSize);
						std::vector<sf::Vector2f> navpath = navmesh.getPath(from, to, polypath, navAgentSize);

						connections.clear();
						path.clear();
						prios.resize(polys.size());

						for(unsigned i = 0; i < polypath.size(); i++) {
							NavmeshExposer::PathNode node = polypath[i];
							const std::vector<sf::Vector2f> &poly = polys[node.poly];
							sf::Vector2f shapeMidpoint(0,0);
							for(int k = 0; k < poly.size(); k++) {
								shapeMidpoint.x += (poly[k].x - shapeMidpoint.x) / (k+1);
								shapeMidpoint.y += (poly[k].y - shapeMidpoint.y) / (k+1);
							}
							sf::Color shapeColor = shapes[i].getFillColor();
							sf::Color color = sf::Color(255 - shapeColor.r, 255
									- shapeColor.g, 255 - shapeColor.b);

							if (i > 0) {
								sf::Vector2f p1 = poly[node.prev_portal];
								sf::Vector2f p2 = poly[(node.prev_portal+1)%poly.size()];
								sf::Vector2f lineMidpoint = sf::Vector2f((p1.x + p2.x)/2.0f, (p1.y + p2.y)/2.0f);
								sf::Vertex v1(lineMidpoint,color);
								sf::Vertex v2(shapeMidpoint,color);
								connections.append(v1);
								connections.append(v2);
							}
							if (i < polypath.size()-1) {
								sf::Vector2f p1 = poly[node.next_portal];
								sf::Vector2f p2 = poly[(node.next_portal+1)%poly.size()];
								sf::Vector2f lineMidpoint = sf::Vector2f((p1.x + p2.x)/2.0f, (p1.y + p2.y)/2.0f);
								sf::Vertex v1(shapeMidpoint,color);
								sf::Vertex v2(lineMidpoint,color);
								connections.append(v1);
								connections.append(v2);
							}
						}

						for(unsigned i = 1; i < navpath.size(); i++) {
							sf::Color color(0, 255, 0);
							sf::Vertex v1, v2;
							v1.color = color;
							v2.color = color;
							v1.position = navpath[i-1];
							v2.position = navpath[i];
							path.append(v1);
							path.append(v2);
						}
						fromSet = false;
					}
				}
			}
		}

		if(up) {
			cameraMove.y -= 1.0f * deltaTime.asSeconds();
		}
		if(down) {
			cameraMove.y += 1.0f * deltaTime.asSeconds();
		}
		if(left) {
			cameraMove.x -= 1.0f * deltaTime.asSeconds();
		}
		if(right) {
			cameraMove.x += 1.0f * deltaTime.asSeconds();
		}

		view.move(normalize(cameraMove));

		window.clear();
		window.setView(view);

		tilemap.drawLayers(window, sf::RenderStates(), 0, -1);

		if(drawLines) {
			window.draw(lines);
		}
		
		if(drawTris) {
			for(int i = 0; i < shapes.size(); i++) {
				window.draw(shapes[i]);
			}
		}
		
		if(drawConnections) {
			window.draw(connections);
			window.draw(path);
			for(int i = 0; i < prios.size(); i++) {
				window.draw(prios[i]);
			}
		}

		window.display();
	}
}
